#include "MainMenuBar.h"
#include <nfd.h>
#include "AssetManagement/AssetManager.h"



MainMenuBar::MainMenuBar()
{
    ImGui::BeginMainMenuBar(); 
}

void MainMenuBar::ApplicationMenu(Project& project) {

    if (ImGui::BeginMenu("Application")) {

        if (ImGui::MenuItem("Load Project"))
        {
            nfdresult_t result = NFD_OpenDialog({ "yproj" }, NULL, &path);
            switch (result) {
            case(NFD_OKAY):
                Project::LoadProject(path, project);
                AssetManager::setAssetPath(project.GetProjectDirectory());
                AssetManager::BuildAssetView();
                break;
            case(NFD_CANCEL):
                break;
            case(NFD_ERROR):
                std::cout << "NFD_Error: " << NFD_GetError() << std::endl;
                break;
            }

        }

        if (ImGui::MenuItem("Save project as...")) {
            nfdresult_t result = NFD_SaveDialog({ "yproj" }, NULL, &path);
            switch (result) {
            case(NFD_OKAY):
                std::cout << "Save as: " << path << std::endl;
                Project::SaveProject(path, project);
                break;
            case(NFD_CANCEL):
                break;
            case(NFD_ERROR):
                std::cout << "NFD_Error: " << NFD_GetError() << std::endl;
                break;
            }
        }


  



        if (ImGui::MenuItem("Preferences"))
        {

        }

        if (ImGui::MenuItem("Exit"))
        {
            // TODO: Exit application
        }

        ImGui::EndMenu();
    }
}

void MainMenuBar::SceneMenu(Project& project, YoggieEngine::Scene& scene) {

    if (ImGui::BeginMenu("Scene")) {

        if (ImGui::MenuItem("Save scene"))
        {
            nfdresult_t result = NFD_SaveDialog({ "yscene" }, NULL, &path);
            switch (result) {
            case(NFD_OKAY):
                SaveScene(path, scene);
                
                project.AddScene(scene);

                break;
            case(NFD_CANCEL):
                break;
            case(NFD_ERROR):
                std::cout << "NFD_Error: " << NFD_GetError() << std::endl;
                break;
            }

        }

        if (ImGui::MenuItem("Load scene"))
        {
            auto result = NFD_OpenDialog({ "yscene" }, NULL, &path);
            switch (result) {
            case (NFD_OKAY):
                LoadScene(path, scene);
                break;
            case(NFD_CANCEL):
                break;
            case(NFD_ERROR):
                std::cout << "NFD_Error: " << NFD_GetError() << std::endl;
                break;
            }

        }

        if (ImGui::MenuItem("Add Entity"))
        {
            scene.AddEntity("New Entity");
        }

        if (ImGui::MenuItem("Import Model"))
        {
            auto result = NFD_OpenDialog("obj,fbx,gltf", NULL, &path);
            switch (result) {
            case(NFD_OKAY):
                // Import Model

                AssetManager::LoadFromSource(
                    path,
                    "build/Debug/Assets"//project.get()->GetProjectDirectory() / "Assets"
                );
                break;
            case(NFD_CANCEL):
                break;
            case(NFD_ERROR):
                std::cout << "NFD_Error: " << NFD_GetError() << std::endl;
                break;
            }
        }

        if (ImGui::MenuItem("Import MeshAsset (temp)"))
        {
            auto result = NFD_OpenDialog("mesh", NULL, &path);

            switch (result) {
            case(NFD_OKAY):
            {
                YoggieEngine::Mesh* importedMesh = AssetManager::LoadFromAssetFile(path);
                if (importedMesh != nullptr)
                {
                    auto full_name = std::filesystem::path(path);
                    auto importedModel = scene.AddEntity(full_name.filename().u8string());
                    auto& rendererComponent = importedModel.AddComponent<YoggieEngine::Render3DComponent>();
                    rendererComponent.mesh = *importedMesh;

                }
            }
            break;

            case(NFD_CANCEL):
                spdlog::debug("User cancelled action");
                break;
            case(NFD_ERROR):
                spdlog::warn("Something went wrong!");
                break;
            }

        }

        ImGui::EndMenu();
    }
}

void MainMenuBar::DebugMenu()
{
    if (ImGui::BeginMenu("Debug")) {
        ImGui::EndMenu();
    }
}

void MainMenuBar::SelectMenu() {
    if (ImGui::BeginMenu("Select")) {
        ImGui::EndMenu();
    }
}

void MainMenuBar::WindowMenu() {
    if (ImGui::BeginMenu("Window")) {

        ImGui::EndMenu();
    }
}

void MainMenuBar::Help() {
    if (ImGui::BeginMenu("Help")) {
        ImGui::EndMenu();
    }
}

MainMenuBar::~MainMenuBar() 
{ 
    ImGui::EndMainMenuBar(); 
}