#include "Viewport.h"

Viewport::Viewport(YoggieEngine::Scene& scene ):
    EditorWindow("SceneView"),
    renderer(YoggieEngine::RendererConfig{ 1200, 700, glm::vec3(0), true }),
    cam(glm::vec3(14.0f, 1.0f, 0.0f), glm::vec3(0.0f, 0.0f, 0.0f), 90),
    scene(scene)
{
    renderer.SetMainCamera(cam);

}

void Viewport::Draw() {

    auto group = scene.getReg().view<YoggieEngine::TransformComponent, YoggieEngine::Render3DComponent>();
    group.each([&](auto enity, YoggieEngine::TransformComponent& t, YoggieEngine::Render3DComponent& renderComponent) {
        renderer.Submit(renderComponent, t);
        });

    cam.Update();
    renderer.SetMainCamera(cam);
    renderer.Render(scene);

    ImVec2 WinPos = ImGui::GetWindowPos();
    ImVec2 ContentRegionMin = ImGui::GetWindowContentRegionMin();
    ImVec2 ContentRegionMax = ImGui::GetWindowContentRegionMax();
    ImVec2 ScreenSpaceMin = { ContentRegionMin.x + WinPos.x, ContentRegionMin.y + WinPos.y };
    ImVec2 ScreenSpaceMax = { ContentRegionMax.x + WinPos.x,ContentRegionMax.y + WinPos.y };


    ImGui::Image(
        (void*)(intptr_t)renderer.getCurrentFrameBuffer().GetColourAttachment(),
        ImVec2{ (float)ImGui::GetWindowWidth(),(float)ImGui::GetWindowHeight() }
    );


    ImGuizmo::Enable(true);
    ImGuizmo::SetRect(ScreenSpaceMin.x, ScreenSpaceMin.y, ContentRegionMax.x, ContentRegionMax.y);

    glm::mat4 transposed_view = glm::transpose(cam.ViewMatrix);
    isFocused = ImGui::IsWindowFocused();

    //ImGuizmo::DrawGrid(glm::value_ptr(cam.ViewMatrix), glm::value_ptr(cam.ProjectionMatrix), glm::value_ptr(worldOrigin), 100.0f);
    //ImGuizmo::ViewManipulate(glm::value_ptr(cam.ViewMatrix), 1, { ScreenSpaceMin.x,ScreenSpaceMin.y }, { 90,90 }, 0x22CCCCCC);

    // Matrix is the model matrix we would want to manipulate
    //ImGuizmo::Manipulate(glm::value_ptr(cam.ViewMatrix), glm::value_ptr(cam.ProjectionMatrix), ImGuizmo::TRANSLATE, ImGuizmo::WORLD, glm::value_ptr(cam.ViewMatrix));

}