#include <YoggieEngine.h>

#include "Renderer.h"
#include "../Scene/Components.h"
#include "../Graphics/Memory/Buffer.h"
#include "../Graphics/Memory/VertexArray.h"
#include "../Graphics/Primitives/DrawCommand.h"

namespace YoggieEngine {
unsigned int quadVAO = 0;
unsigned int quadVBO = 0;
// vegetation test
std::vector<glm::vec3> vegetation = {
	glm::vec3(-1.5f, 0.0f, -0.48f),
	glm::vec3(1.5f, 0.0f, 0.51f),
	glm::vec3(0.0f, 0.0f, 0.7f),
	glm::vec3(-0.3f, 0.0f, -2.3f)
};

unsigned int transparentVAO, transparentVBO;
Texture grassTexture;

float skyboxVertices[]{
	// positions          
	-1.0f,  1.0f, -1.0f,
	-1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,

	-1.0f, -1.0f,  1.0f,
	-1.0f, -1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f,  1.0f,
	-1.0f, -1.0f,  1.0f,

	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,

	-1.0f, -1.0f,  1.0f,
	-1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f, -1.0f,  1.0f,
	-1.0f, -1.0f,  1.0f,

	-1.0f,  1.0f, -1.0f,
	 1.0f,  1.0f, -1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	-1.0f,  1.0f,  1.0f,
	-1.0f,  1.0f, -1.0f,

	-1.0f, -1.0f, -1.0f,
	-1.0f, -1.0f,  1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	-1.0f, -1.0f,  1.0f,
	 1.0f, -1.0f,  1.0f
};

Renderer::Renderer(RendererConfig& config)
	: m_framebuffer(Framebuffer(config.ScreenWidth, config.ScreenHeight)),
	gBufferShader("build/Debug/Shaders/deferred/geometry.vert", "build/Debug/Shaders/deferred/geometry.frag"),
	lightingPassShader("build/Debug/Shaders/deferred/lightPass.vert", "build/Debug/Shaders/deferred/lightPass.frag"),
	SkyboxShader("build/Debug/Shaders/Cubemaps/Skybox.vert", "build/Debug/Shaders/Cubemaps/Skybox.frag"),
	BlendingShader("build/Debug/Shaders/forward/Blending.vert", "build/Debug/Shaders/forward/Blending.frag"),
	forwardShader("build/Debug/Shaders/forward/geometry.vert", "build/Debug/Shaders/forward/geometry.frag"),
	postProcessingShader("build/Debug/Shaders/forward/postprocessing.vert", "build/Debug/Shaders/forward/postprocessing.frag")
{	
	width = config.ScreenWidth;
	height = config.ScreenHeight;
	

	glEnable(GL_DEPTH_TEST);
	// Deferred Rendering 
	glGenFramebuffers(1, &gBuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);
	
	// - Position Color Buffer ;
	glGenTextures(1, &gPosition);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gPosition, 0);

	// - normal color buffer ;
	glGenTextures(1, &gNormal);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, gNormal, 0);

	// - Color + Specular buffer ;
	glGenTextures(1, &gColorSpec);
	glBindTexture(GL_TEXTURE_2D, gColorSpec);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, gColorSpec, 0);


	unsigned int attachments[3] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 , GL_COLOR_ATTACHMENT2 };
	glDrawBuffers(3, attachments);

	// Create and attach a depth buffer 
	glGenRenderbuffers(1, &gDepth);
	glBindRenderbuffer(GL_RENDERBUFFER, gDepth);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, gDepth);

	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
	{
		spdlog::critical("Framebuffer not complete {}{}", __FILE__, __LINE__);
	}

	

	std::vector<std::string> faces{
	"build/Debug/skybox/Open_Water/right.jpg",
	"build/Debug/skybox/Open_Water/left.jpg",
	"build/Debug/skybox/Open_Water/top.jpg",
	"build/Debug/skybox/Open_Water/bottom.jpg",
	"build/Debug/skybox/Open_Water/front.jpg",
	"build/Debug/skybox/Open_Water/back.jpg"
	};

	sky = CubeMap(faces);

	// Create a skybox vao
	unsigned int VBO;
	glGenVertexArrays(1, &skyboxVAO);
	glGenBuffers(1, &VBO);
	glBindVertexArray(skyboxVAO);
	glBindBuffer(GL_ARRAY_BUFFER, VBO);
	glBufferData(GL_ARRAY_BUFFER, sizeof(skyboxVertices), &skyboxVertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);

#ifdef WINDOW 
	grassTexture = Texture("build/Debug/Texture/blending_transparent_window.png", true);
#else 
	grassTexture = Texture("build/Debug/Texture/grass.png", true);
#endif

}

Renderer::~Renderer(){}

void SubmitVegetationDemo() {


	if (transparentVAO == 0) {
		float transparentVertices[] = {
			// positions         // texture Coords (swapped y coordinates because texture is flipped upside down)
			0.0f,  0.5f,  0.0f,  0.0f,  0.0f,
			0.0f, -0.5f,  0.0f,  0.0f,  1.0f,
			1.0f, -0.5f,  0.0f,  1.0f,  1.0f,

			0.0f,  0.5f,  0.0f,  0.0f,  0.0f,
			1.0f, -0.5f,  0.0f,  1.0f,  1.0f,
			1.0f,  0.5f,  0.0f,  1.0f,  0.0f
		};

		glGenVertexArrays(1, &transparentVAO);
		glGenBuffers(1, &transparentVBO);
		glBindVertexArray(transparentVAO);
		glBindBuffer(GL_ARRAY_BUFFER, transparentVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(transparentVertices), transparentVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
		glBindVertexArray(0);

	}


}

void Renderer::Submit( Render3DComponent& renderComponent, TransformComponent& transform) {
	if (renderComponent.VAO == 0 || renderComponent.IBO == 0) 
	{
		if (renderComponent.VAO != 0)
			glDeleteVertexArrays(1, &(renderComponent.VAO));
		if (renderComponent.IBO != 0)
			glDeleteBuffers(1, &(renderComponent.IBO));

		VertexArray va = VertexArray();
		Buffer vertexBuffer = Buffer();
		Buffer elementBuffer = Buffer();

		va.Create();
		va.Bind();

		vertexBuffer.createBuffer();
		vertexBuffer.Bind(false);
		vertexBuffer.setBufferData((void*)&renderComponent.mesh.vertices[0], renderComponent.mesh.vertices.size() * sizeof(Vertex), false);

		elementBuffer.createBuffer();
		elementBuffer.Bind(true);
		elementBuffer.setBufferData((void*)&renderComponent.mesh.elements[0], renderComponent.mesh.elements.size() * sizeof(unsigned int), true);

		va.AttachAttribute(0, 3, sizeof(Vertex));

		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (void*)0);
		glEnableVertexAttribArray(0);
		glEnableVertexAttribArray(1);

		va.Unbind();
		vertexBuffer.Unbind(false);
		elementBuffer.Unbind(true);

		renderComponent.VAO = va.getID();
		renderComponent.IBO = elementBuffer.getBufferID();
	}

	DrawCommand dc = { renderComponent.isStatic , renderComponent.VAO, renderComponent.mesh.elements.size(), renderComponent.IBO, transform, renderComponent.color };
	commands.push_back(dc);
	
}

void Renderer::GeometryPass() {
	// 1.0 Geometry pass 
	gBufferShader.Use();

	for (const DrawCommand& command : commands)
	{
		if (command.isDynamic == true)
			continue;

		glBindVertexArray(command.VAO_identifier);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, command.IBO_identifier);

		gBufferShader.setUniformVec3("Color", command.color);
		gBufferShader.setUniformMat4("Model", command.transform.LocalTransform);
		gBufferShader.setUniformMat4("View", MainCamera.ViewMatrix);
		gBufferShader.setUniformMat4("Projection", MainCamera.ProjectionMatrix);

		glDrawElements(GL_TRIANGLES, static_cast<unsigned int>(command.num_elements),
			GL_UNSIGNED_INT, NULL);

		glBindVertexArray(0);
	}


}

void Renderer::ForwardGeometryPass() 
{
	for (const DrawCommand& command : commands)
	{
		if (command.isDynamic == false)
			continue;

		glBindVertexArray(command.VAO_identifier);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, command.IBO_identifier);

		forwardShader.Use();

		forwardShader.setUniformVec3("Color", command.color);
		forwardShader.setUniformMat4("M", command.transform.LocalTransform);
		forwardShader.setUniformMat4("V", MainCamera.ViewMatrix);
		forwardShader.setUniformMat4("P", MainCamera.ProjectionMatrix);

		glDrawElements(GL_TRIANGLES, static_cast<unsigned int>(command.num_elements),
			GL_UNSIGNED_INT, NULL);

		glBindVertexArray(0);

	}

}

void Renderer::SkyboxPass() {
	// Render skybox 
	glDepthMask(GL_FALSE);
	SkyboxShader.Use();
	SkyboxShader.setUniformMat4("projection", MainCamera.ProjectionMatrix);
	SkyboxShader.setUniformMat4("view", glm::mat4(glm::mat3(MainCamera.ViewMatrix))); // remove rotation from the view matrix
	glBindVertexArray(skyboxVAO);
	glBindTexture(GL_TEXTURE_CUBE_MAP, sky.getID());
	glDrawArrays(GL_TRIANGLES, 0, 36);
	glBindVertexArray(0);
	glDepthMask(GL_TRUE);
	
}

void Renderer::lightingPass(Scene& scene){

	// 2.0 Lighting Pass

	// configure shader 
	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer.GetId());

	lightingPassShader.Use();
	lightingPassShader.setUniformInt("gPosition", 0);
	lightingPassShader.setUniformInt("gNormal", 1);
	lightingPassShader.setUniformInt("gColorSpec", 2);

	// Bind all Gbuffer textures
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, gColorSpec);

	//	send relevant Lighting Uniforms
	auto lights = scene.getReg().view<LightComponent, TransformComponent>();
	unsigned int lightnr = 0;
	lights.each([&](auto entity, LightComponent& light, TransformComponent& transform) {
		lightingPassShader.setUniformVec3("lights[" + std::to_string(lightnr) + "].Position", transform.Position);
		lightingPassShader.setUniformVec3("lights[" + std::to_string(lightnr) + "].Color", light.Color);

		// Attenuation
		const float linear = 0.7f;
		const float quadratic = 1.8f;
		lightingPassShader.setUniformFloat("lights[" + std::to_string(lightnr) + "].Linear", linear);
		lightingPassShader.setUniformFloat("lights[" + std::to_string(lightnr) + "].Quadratic", quadratic);


		lightnr++;
	});

	lightingPassShader.setUniformVec3("viewPos", MainCamera.Position);

	// render to quad
	if (quadVAO == 0)
	{
		float quadVertices[] = {
			-1.0f,	 1.0f, 0.0f, 0.0f, 1.0f,
			-1.0f,	-1.0f, 0.0f, 0.0f, 0.0f,
			 1.0f,	 1.0f, 0.0f, 1.0f, 1.0f,
			 1.0f,	-1.0f, 0.0f, 1.0f, 0.0f,
		};
		// setup plane VAO ;
		glGenVertexArrays(1, &quadVAO);
		glGenBuffers(1, &quadVBO);
		glBindVertexArray(quadVAO);
		glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
	}


	glBindVertexArray(quadVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glBindVertexArray(0);
}

void Renderer::BlendingPass() {
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Vegetation / blending test;
	BlendingShader.Use();
	glBindVertexArray(transparentVAO);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, grassTexture.GetID());

	BlendingShader.setUniformMat4("V", MainCamera.ViewMatrix);
	BlendingShader.setUniformMat4("P", MainCamera.ProjectionMatrix);


	for (unsigned int i = 0; i < vegetation.size(); i++) {
		auto rotation = glm::rotate(glm::mat4(1.0f), 45.0f, glm::vec3(0.0f, 1.0f, 0.0f));
		auto translation = glm::translate(glm::mat4(1.0f), vegetation[i]);
		auto transform = translation * rotation;
		BlendingShader.setUniformMat4("M", transform);
		glDrawArrays(GL_TRIANGLES, 0, 6);

	}

	glBindVertexArray(0);

	glDisable(GL_BLEND);



}

void Renderer::PostProcessing() 
{
	postProcessingShader.Use();
	postProcessingShader.setUniformInt("screen", 0);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, m_framebuffer.GetColourAttachment());


	glBindVertexArray(quadVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glBindVertexArray(0);

	
}

void Renderer::CopyGBuffer() {
	glBindFramebuffer(GL_READ_FRAMEBUFFER, gBuffer);
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, m_framebuffer.GetId());
	glBlitFramebuffer(0, 0, width, height, 0, 0, width, height, GL_DEPTH_BUFFER_BIT, GL_NEAREST);

	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer.GetId());
}

void Renderer::Render(Scene& scene) 
{	
	SubmitVegetationDemo();
	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	SkyboxPass();

	GeometryPass();

	lightingPass(scene);
	
	CopyGBuffer();

	ForwardGeometryPass();

	BlendingPass();

	//PostProcessing();


	//  Lighting pass 
	/*
	auto lights = scene.getReg().view<LightComponent>();
		lights.each([&](auto entity, LightComponent& light) {
		  	renderComponent.shader.setUniformVec3("lighting.color", light.Color);
			renderComponent.shader.setUniformFloat("lighting.strength", light.Strength);
		});
	
	*/ 	
	
	commands.clear();
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void Renderer::setCurrentFrameBuffer(const Framebuffer& fb)
{
	m_framebuffer = fb;
}

void Renderer::setClearColor(const glm::vec3& ClearColor)
{
	m_clearColor = ClearColor;
}

}